#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "./modeldef.h"

#define TEST 0

extern void plmbar_(double *p,double *dp,double *costh,int *degree,int *inorm);
extern void magfdz_(double *p,double *dp,double *theta,double *phi,double *rad,
                    int *degree,double *g,double *dx,double *dy,double *dz,
                    double *dpot,double *x,double *y,double *z,double *h,
                    double *f,double *ainc,double *d,double *pot);
extern void magfdz_ext_(double *p,double *dp,double *theta,double *phi,double *rad,
                    int *degree,double *g,double *dx,double *dy,double *dz,
                    double *dpot,double *x,double *y,double *z,double *h,
                    double *f,double *ainc,double *d,double *pot);

extern int sm2geo_init(char sm2geo_fname[],int *deg_ext,int *deg_day,int maxcoeff,double coeff[SM2GEO_MAXNCOEFF]);

extern void sm2geo(double fday,int deg_wanted,                                          /* chosen */
                   int deg_ext,int deg_day,double coeff[SM2GEO_MAXNCOEFF],              /* given */
                   double smcoeff[MAX_EXT_NCOEFF],double geocoeff[MAX_EXT_NCOEFF]);

extern void sm2geo_green(double fday,int deg_want,                               /* chosen */
                         int deg_ext,int deg_day,double coeff[SM2GEO_MAXNCOEFF], /* given */
                         tgsm2geo_green green_day_r,tgsm2geo_green green_day_i);

extern int gsm2geo_init(char gsm2geo_fname[],int *deg_ext,int *deg_day,int *deg_ann,int maxcoeff,double coeff[GSM2GEO_MAXNCOEFF]);

extern void gsm2geo(double fday,int deg_wanted,                                               /* chosen */
                    int deg_ext,int deg_day,int deg_ann,double coeff[GSM2GEO_MAXNCOEFF],      /* given */
                    double gsmcoeff[MAX_EXT_NCOEFF],double geocoeff[MAX_EXT_NCOEFF]);

extern void gsm2geo_green(double fday,int deg_want,                                           /* chosen */
                          int deg_ext,int deg_day,int deg_ann,double coeff[SM2GEO_MAXNCOEFF], /* given */
                          tgsm2geo_green green_day_r,tgsm2geo_green green_day_i,
                          tgsm2geo_green green_ann_r,tgsm2geo_green green_ann_i);

extern void geodetic2geocentric(double phi, double h,double *latrad, double *r); /* phi is geodetic latitude in radian*/

extern void geocentric2geodetic_vec(double theta, double delta, double bx, double by, double bz,
                                    double *x, double *y, double *z);

static double dmin(double x,double y){ if (x < y) return x; else return y; }
static double dmax(double x,double y){ if (x > y) return x; else return y; }

static FILE *infile;
static void readln(char oneline[256])
{
  char *temp;
  fgets(oneline, 256, infile);  
  temp = strchr(oneline, '\n');  
  if (temp != NULL) *temp = 0;
/*    printf("%s\n",oneline); */
/*    if (temp>=0&&temp<256) oneline[temp] = 0; */
  return;
}

static int indx(int n,int m)
{
  int i;

  if (m == 0) {
    i = (n - 1) * (n + 1) + 1;
    return i-1;
  }
  if (m >= 0)
    i = (n - 1) * (n + 1) + m * 2;
  else
    i = (n - 1) * (n + 1) - m * 2 + 1;
  return i-1;
}  /*indx*/


void mainfield_open_file_mp(char mainfname[],char sm2geo_fname[],char gsm2geo_fname[],int control[MAXCONTROL],int ncontrol, tmodel *model, int model_nbytes)
{
  char oneline[256]=" ",code[7];
  double g, h, gdot, hdot,gdotdot,hdotdot,gdotdot_plus,hdotdot_plus,epoch,data_start,data_end;
  int i,i1,i2,iline,m,n,status,gsm2geo_ncoeff,sm2geo_ncoeff;

  (*model).first_day = -100000.0; /* in case the date range is not given in the file */
  (*model).last_day = 100000.0;


  if (model_nbytes < sizeof(tmodel))
    {printf("Model array needs at least %1d bytes, while caller provides %1d\n",(int)sizeof(tmodel),model_nbytes);exit(EXIT_FAILURE);}

  if (ncontrol < MAXCONTROL)
    {printf("Model control array needs at least %1d elements\n",(int) MAXCONTROL);exit(EXIT_FAILURE);}

  memset(model, 0, sizeof(tmodel));

  printf("reading %s...", mainfname);fflush(stdout);
  if (infile != NULL)
    infile = freopen(mainfname, "r", infile);
  else
    infile = fopen(mainfname, "r");
  if (infile == NULL) {printf("%s%s%s\n","file ",mainfname," could not be opened");;exit(EXIT_FAILURE);}
  
next_header:
  readln(oneline); 
  if (oneline[0]=='#') goto next_header;

  sscanf(oneline,"%6s",code);
  if (TEST) printf("next code: %s\n",code);

  if (strcmp(code,"Epoch:") == 0) 
    {
      fscanf(infile, "%lg%*[^\n]",&epoch);
      getc(infile);
      if (TEST) printf("%s%8.2f\n",oneline,epoch);
      (*model).ref_day = (epoch - 2000.0) * 365.25;
      if ((epoch < 2000) || (epoch > 2010))
        {printf("unknown epoch %f\n",epoch); exit(EXIT_FAILURE);}
      goto next_header;
    }
  if (strcmp(code,"Data-r") == 0) 
    {
      fscanf(infile, "%lg%lg%*[^\n]",&data_start,&data_end);
      getc(infile);
      if (TEST) printf("%sange: data_start=%.2f, data_end=%.2f\n",oneline,data_start,data_end);
      (*model).first_day = (data_start - 2000.0) * 365.25;
      (*model).last_day = (data_end - 2000.0) * 365.25;
      if ((dmin(data_start,data_end) < 1999.0) || (dmax(data_start,data_end) > 2020.0))
        {printf("Please check data interval: start_date=%f, end_dat=%f\n",data_start,data_end); exit(EXIT_FAILURE);}
      goto next_header;
    }
  if (strcmp(code,"IMF-By") == 0) 
    {
      fscanf(infile, "%lg%lg%*[^\n]",&(*model).imfby[1],&(*model).imfby[2]);  /* imfby[0] stays empty */
      getc(infile);
      if (TEST) printf("%s %8.3f %8.3f\n",oneline,(*model).imfby[1],(*model).imfby[2]);
      control[POS_IMFBY] = 1;
      control[POS_IMFBY_IND] = 1;
      (*model).deg_imfby = 1;
      goto next_header;
    }
  if (strcmp(code,"SM-Fac") == 0) 
    {
      fscanf(infile, "%lg%*[^\n]",&(*model).est_fac);
      getc(infile);
      if (TEST) printf("%s %8.4f\n",oneline, (*model).est_fac);
      control[POS_EST_FAC] = 1;
      goto next_header;
    }
  if (strcmp(code,"SM-Ext") == 0) 
    {
      fscanf(infile, "%d%*[^\n]",&(*model).deg_sm);
      getc(infile);
      if (TEST) printf("%s degree %d\n",oneline,(*model).deg_sm);
      if ((*model).deg_sm > 0)
        { 
          fscanf(infile, "%lg%lg%lg%*[^\n]",&(*model).sm[0],&(*model).sm[1],&(*model).sm[2]); 
          getc(infile);
          if (TEST) printf("%10.3f %10.3f %10.3f \n",(*model).sm[0],(*model).sm[1],(*model).sm[2]); 
        }
      if ((*model).deg_sm > 1)
        {
          fscanf(infile, "%lg%lg%lg%lg%lg%*[^\n]",&(*model).sm[3],&(*model).sm[4],&(*model).sm[5],&(*model).sm[6],&(*model).sm[7]);
          getc(infile);
          if (TEST) 
            printf("%10.3f %10.3f %10.3f %10.3f %10.3f \n",(*model).sm[3],(*model).sm[4],(*model).sm[5],(*model).sm[6],(*model).sm[7]);
        }
      if ((*model).deg_sm > 2)
        {
          fscanf(infile, "%lg%lg%lg%lg%lg%lg%lg%*[^\n]",
                 &(*model).sm[8],&(*model).sm[9],&(*model).sm[10],&(*model).sm[11],&(*model).sm[12],&(*model).sm[13],&(*model).sm[14]);
          getc(infile);
          if (TEST) 
            printf("%10.3f %10.3f %10.3f %10.3f %10.3f %10.3f %10.3f \n",
                   (*model).sm[8],(*model).sm[9],(*model).sm[10],(*model).sm[11],(*model).sm[12],(*model).sm[13],(*model).sm[14]);
        }
      control[POS_SM] = (*model).deg_sm;
      control[POS_SM_IND] = (*model).deg_sm;
      goto next_header;
    }
  if (strcmp(code,"GSM-Ex") == 0) 
    {
      fscanf(infile, "%d%*[^\n]",&(*model).deg_gsm);
      getc(infile);
      if (TEST) printf("%s degree %d\n",oneline,(*model).deg_gsm);
      if ((*model).deg_gsm > 0)
        { 
          fscanf(infile, "%lg%lg%lg%*[^\n]",&(*model).gsm[0],&(*model).gsm[1],&(*model).gsm[2]); 
          getc(infile);
          if (TEST) printf("%10.3f %10.3f %10.3f \n",(*model).gsm[0],(*model).gsm[1],(*model).gsm[2]); 
        }
      if ((*model).deg_gsm > 1)
        {
          fscanf(infile, "%lg%lg%lg%lg%lg%*[^\n]",&(*model).gsm[3],&(*model).gsm[4],&(*model).gsm[5],
                 &(*model).gsm[6],&(*model).gsm[7]);
          getc(infile);
          if (TEST) 
            printf("%10.3f %10.3f %10.3f %10.3f %10.3f \n",(*model).gsm[3],(*model).gsm[4],(*model).gsm[5],
                 (*model).gsm[6],(*model).gsm[7]);
        }
      if ((*model).deg_gsm > 2)
        {
          fscanf(infile, "%lg%lg%lg%lg%lg%lg%lg%*[^\n]",&(*model).gsm[8],&(*model).gsm[9],
                 &(*model).gsm[10],&(*model).gsm[11],&(*model).gsm[12],&(*model).gsm[13],&(*model).gsm[14]);
          getc(infile);
          if (TEST) 
            printf("%10.3f %10.3f %10.3f %10.3f %10.3f %10.3f %10.3f \n",(*model).gsm[8],(*model).gsm[9],
                 (*model).gsm[10],(*model).gsm[11],(*model).gsm[12],(*model).gsm[13],(*model).gsm[14]);
        }
      control[POS_GSM] = (*model).deg_gsm;
      control[POS_GSM_IND] = (*model).deg_gsm;
      goto next_header;
    }
  if (strcmp(code,"Intern") == 0) 
    {
      readln(oneline);/*  Degree: static 1st-deriv 2nd-deriv-before 2nd-deriv-after */
      fscanf(infile, "%d%d%d%d%*[^\n]",&(*model).deg_int,&(*model).deg_1st,&(*model).deg_2nd,&(*model).deg_2nd_plus);
      control[POS_0] = (*model).deg_int;
      control[POS_1] = (*model).deg_1st;
      control[POS_2] = (*model).deg_2nd;
      control[POS_3] = (*model).deg_2nd_plus;
      getc(infile);
      if (TEST) printf("%s%6d%6d%6d%6d\n",oneline,(*model).deg_int,(*model).deg_1st,(*model).deg_2nd,(*model).deg_2nd_plus);
      if ((*model).deg_int > MAX_0_DEG)
        {printf("To use this model increase the parameter MAX_0_DEG from %3d to %3d \n",(int)MAX_0_DEG,(*model).deg_int);exit(EXIT_FAILURE);}
      if ((*model).deg_1st > MAX_V_DEG)
        {printf("To use this model increase the parameter MAX_V_DEG from %3d to %3d \n",(int)MAX_V_DEG,(*model).deg_1st);exit(EXIT_FAILURE);}
      if ((*model).deg_2nd > MAX_V_DEG)
        {printf("To use this model increase the parameter MAX_V_DEG from %3d to %3d \n",(int)MAX_V_DEG,(*model).deg_2nd);exit(EXIT_FAILURE);}
      if ((*model).deg_2nd_plus > MAX_V_DEG)
        {printf("To use this model increase the parameter MAX_V_DEG from %3d to %3d \n",(int)MAX_V_DEG,(*model).deg_2nd_plus);exit(EXIT_FAILURE);}

      readln(oneline);/*  n m glm hlm dglm/dt dhlm/dt ... */
      readln(oneline);/*  --------------------------- */

      for (i = 1; i <= (*model).deg_int; i++) 
        {
          for (iline = 0; iline <= i; iline++) 
            {
              readln(oneline);
              status = sscanf(oneline, "%d%d", &n, &m);
              if (i != n || iline != m) 
                {
                  printf("Error reading internal field coefficients\n expected n = %1d m = %1d, read n = %1d m = %1d\n",
                         i,iline,n,m);
                  exit(EXIT_FAILURE);
                }
              if (n <= (*model).deg_2nd_plus)
                status = sscanf(oneline, "%d%d%lg%lg%lg%lg%lg%lg%lg%lg", &n, &m, &g, &h, &gdot,&hdot, 
                                &gdotdot,&hdotdot,&gdotdot_plus,&hdotdot_plus);
              else
                if (n <= (*model).deg_2nd)
                  status = sscanf(oneline, "%d%d%lg%lg%lg%lg%lg%lg", &n, &m, &g, &h, &gdot,&hdot, 
                                  &gdotdot,&hdotdot);
                else
                  if (n <= (*model).deg_1st)
                    status = sscanf(oneline, "%d%d%lg%lg%lg%lg", &n, &m, &g, &h, &gdot,&hdot);
                  else
                    status = sscanf(oneline, "%d%d%lg%lg", &n, &m, &g, &h);
              
              i1 = indx(n, m);
              i2 = indx(n, -m);
              (*model).g[i1] = g;
              if (n <= (*model).deg_1st) (*model).gs[i1] = gdot;
              if (n <= (*model).deg_2nd) (*model).gssm[i1] = gdotdot;
              if (n <= (*model).deg_2nd_plus) (*model).gssp[i1] = gdotdot_plus;
              if (n > (*model).deg_2nd_plus && n <= (*model).deg_2nd) (*model).gssp[i1] = gdotdot; /* if no split in SA */
              if (m > 0) 
                {
                  (*model).g[i2] = h;
                  if (n <= (*model).deg_1st) (*model).gs[i2] = hdot; 
                  if (n <= (*model).deg_2nd) (*model).gssm[i2] = hdotdot;
                  if (n <= (*model).deg_2nd_plus) (*model).gssp[i2] = hdotdot_plus;
                  if (n > (*model).deg_2nd_plus && n <= (*model).deg_2nd) (*model).gssp[i2] = hdotdot; /* if no split in SA */
                }
            }
        }
      if (infile != NULL)
        fclose(infile);
      infile = NULL;
      if (TEST) 
        if ( (*model).deg_int > 0)
          printf("%s%6d%s%6d%12.4f%12.4f\n","last index n, m: ",n," ",m,
                 (*model).g[indx(n, m)],(*model).g[indx(n, -m)]);
    } /* if code = internal field  */
  else 
    {
      printf(" Pomme model read error !!! \n");
      exit(EXIT_FAILURE);
    }

  printf("done: epoch=%6.1f, n=%1d, n_SV=%1d,  n_SM=%1d,  n_GSM=%1d\n",
         epoch,(*model).deg_int,(*model).deg_1st,(*model).deg_sm,(*model).deg_gsm);
  fflush(stdout);

  sm2geo_ncoeff = sm2geo_init(sm2geo_fname,&(*model).sm2geo_deg_ext,&(*model).sm2geo_deg_day,
                                (int)SM2GEO_MAXNCOEFF,(*model).coeff_sm2geo);
  gsm2geo_ncoeff = gsm2geo_init(gsm2geo_fname,&(*model).gsm2geo_deg_ext,&(*model).gsm2geo_deg_day,&(*model).gsm2geo_deg_ann,
                                (int)GSM2GEO_MAXNCOEFF,(*model).coeff_gsm2geo);
  control[POS_EST] = 1;
  control[POS_IST] = 1;
} /* mainfield_open_file */

void mainfield_get_coeff_mp(double fday,int deg_1st,int deg_2nd,
                     tmodel *model,
                     double g[MAX_0_NCOEFF],double gs[MAX_0_NCOEFF],
                     double gss[MAX_0_NCOEFF])
  {
    double t,t_SA;
    int m,n,i;

  t = (fday-(*model).ref_day)/365.25;
  t_SA = (dmax((*model).first_day,dmin((*model).last_day,fday))-(*model).ref_day)/365.25; /* use SA only within time interval with data coverage*/

  for (n = 1; n <= (*model).deg_int; n++) 
    for (m = -n; m <= n; m++) 
      {
        i = indx(n, m);

        g[i] = (*model).g[i];
        if (n<=deg_1st && n <= MAX_V_DEG) g[i] += t*(*model).gs[i]; 
        if (n<=deg_2nd && n <= MAX_V_DEG)
          {
            if (t <= 0)
              g[i] += 0.5*t_SA*t_SA*(*model).gssm[i] + (t-t_SA)*t_SA*(*model).gssm[i]; /* quadratic back to first_day, linear beyond */
            else 
              g[i] += 0.5*t_SA*t_SA*(*model).gssp[i] + (t-t_SA)*t_SA*(*model).gssp[i]; /* quadratic to last_day, linear beyond */
          }

        gs[i] = (*model).gs[i]; 
        if (n<=deg_2nd && n <= MAX_V_DEG) 
          {
            if (t <= 0)
              gs[i] += t_SA*(*model).gssm[i];
            else 
              gs[i] += t_SA*(*model).gssp[i];
          }

        if (t > t_SA) gss[i] = 0; /* after last data day */
        else 
          if (t > 0) 
            gss[i] = (*model).gssp[i];
          else
            if (t > -t_SA) 
              gss[i] = (*model).gssm[i]; 
            else
              gss[i] = 0;       /* before first data day */
      }
  return;
} /* mainfield_get_coeff */

void mainfield_internal_field_mp (int control[MAXCONTROL], tmodel *model,
                           double fday, double rad, double theta, double phi,
                           double *x, double *y, double *z, double *f) 
{
  double costh,t,t_SA;
  int inorm = 1;	/*1=schmid, 2=fully*/
  double h,ainc,d,pot;
  double g[MAX_0_NCOEFF];
  double p[(MAX_0_DEG+1)*(MAX_0_DEG+2)/2],dp[(MAX_0_DEG+1)*(MAX_0_DEG+2)/2],dx[MAX_0_DEG*(MAX_0_DEG+2)];
  double dy[MAX_0_DEG*(MAX_0_DEG+2)],dz[MAX_0_DEG*(MAX_0_DEG+2)],dpot[MAX_0_DEG*(MAX_0_DEG+2)];
  int i,n,m;

/*    printf("to_deg: %d  fday: %12.2f  rad: %12.2f  theta: %8.2f phi: %8.2f \n",to_deg,fday,rad,theta,phi); */

  t = (fday-(*model).ref_day)/365.25;
  t_SA = (dmax((*model).first_day,dmin((*model).last_day,fday))-(*model).ref_day)/365.25; /* use SA only within time interval with data coverage*/

  if (control[POS_0] > (*model).deg_int) control[POS_0] = (*model).deg_int;

  for (n = 1; n <= control[POS_0]; n++)
    for (m = -n; m <= n; m++) 
      {
        i = indx(n, m);
        g[i] = (*model).g[i];
        if (n <= (*model).deg_1st && n <= MAX_V_DEG) g[i] += t*(*model).gs[i]; 
        if (n <= (*model).deg_2nd && n <= MAX_V_DEG)
          {
            if (t <= 0)
              g[i] += 0.5*t_SA*t_SA*(*model).gssm[i] + (t-t_SA)*t_SA*(*model).gssm[i]; /* quadratic back to first_day, linear beyond */
            else 
              g[i] += 0.5*t_SA*t_SA*(*model).gssp[i] + (t-t_SA)*t_SA*(*model).gssp[i]; /* quadratic to last_day, linear beyond */
          } /* if n within degree of SA */
      } /* for m and n */

  costh = cos(theta);
  plmbar_(&p[0],&dp[0],&costh,&control[POS_0],&inorm);
  magfdz_(&p[0],&dp[0],&theta,&phi,&rad,&control[POS_0],&g[0],
          &dx[0],&dy[0],&dz[0],&dpot[0],x,y,z,&h,f,&ainc,&d,&pot);
  return;
} /* mainfield_internal_field */


void mainfield_external_field_mp (int control[MAXCONTROL], tmodel *model,
                          double fday, double rrel, double theta, double phi,
                          double est, double ist, double imf_by,
                          double *x, double *y, double *z)
{
#include "./qfac.h"
  int igeo,igsm,ism,iperiod,gsm2geo_deg;
  double tx,ty,tz,xi,yi,zi,xe,ye,ze,fac;
  double geo_ext_field[MAX_EXT_NCOEFF],geo_ind_field[MAX_EXT_NCOEFF];
  double induc_sm[MAX_EXT_NCOEFF],induc_gsm[MAX_EXT_NCOEFF],induc_imfby[MAX_EXT_NCOEFF];
  double est_in_geo[MAX_EXT_NCOEFF],ist_in_geo[MAX_EXT_NCOEFF];
  double sm_est_field[MAX_EXT_NCOEFF],sm_ist_field[MAX_EXT_NCOEFF],gsm_imfby_field[MAX_EXT_NCOEFF];
  double sm_in_geo[MAX_EXT_NCOEFF],gsm_in_geo[MAX_EXT_NCOEFF],imfby_in_geo[MAX_EXT_NCOEFF];
  tgsm2geo_green green_day_r, green_ann_r, green_day_i, green_ann_i;
  int ext_deg = MAX_EXT_DEG;
  double costh,t;
  int inorm = 1;	/*1=schmid, 2=fully*/
  double h,ainc,d,pot,f,rad;
  double g[MAX_EXT_NCOEFF];
  double p[(MAX_EXT_DEG+1)*(MAX_EXT_DEG+2)/2],dp[(MAX_EXT_DEG+1)*(MAX_EXT_DEG+2)/2],dx[MAX_EXT_DEG*(MAX_EXT_DEG+2)];
  double dy[MAX_EXT_DEG*(MAX_EXT_DEG+2)],dz[MAX_EXT_DEG*(MAX_EXT_DEG+2)],dpot[MAX_EXT_DEG*(MAX_EXT_DEG+2)];
  int n,mgsm,msm,mgeo;

  if (TEST) printf("1:%f %f %f %f %f %f %f %f %f %f\n",fday,rrel,theta,phi,est,ist,imf_by,*x,*y,*z);
 
  for (igeo=0;igeo<MAX_S_COEFF;igeo++) 
    {
      geo_ext_field[igeo] = 0;
      geo_ind_field[igeo] = 0;
    }

  if (TEST)
    for (igeo=0;igeo<MAX_S_COEFF;igeo++) 
      {
        sm_in_geo[igeo] = 0;
        induc_sm[igeo] = 0;
        est_in_geo[igeo] = 0;
        ist_in_geo[igeo] = 0;
        gsm_in_geo[igeo] = 0;
        induc_gsm[igeo] = 0;
        imfby_in_geo[igeo] = 0;
        induc_imfby[igeo] = 0;
      }

                                /* SM */

  if (control[POS_SM] > 0)
    {
      sm2geo(fday, control[POS_SM], (*model).sm2geo_deg_ext, (*model).sm2geo_deg_day,  (*model).coeff_sm2geo,
             (*model).sm,sm_in_geo);
      for (igeo=0;igeo<control[POS_SM]*(control[POS_SM]+2);igeo++)
        geo_ext_field[igeo] +=  sm_in_geo[igeo];
    }

                                /* Induced by SM in GEO */

  if (control[POS_SM_IND] > 0)
    {
      sm2geo_green(fday, control[POS_SM_IND], (*model).sm2geo_deg_ext, (*model).sm2geo_deg_day, (*model).coeff_sm2geo,
                   green_day_r, green_day_i);
      
      for (igeo=0;igeo<MAX_S_COEFF;igeo++) induc_sm[igeo] = 0;
      
      for (n = 1;n<=control[POS_SM_IND];n++)
        for (msm = -n;msm<=n;msm++)
          {
            ism = indx(n,msm);
            for (mgeo = -n;mgeo<=n;mgeo++)
              {
                igeo = indx(n,mgeo);
                for (iperiod = 1; iperiod <= (*model).sm2geo_deg_day; iperiod++)	      /*green[,,0] is constant term*/
                  induc_sm[igeo] += qfac_day_r[n-1][iperiod - 1] * (*model).sm[ism] * green_day_r[ism][igeo][iperiod]
                    - CONVENTION * qfac_day_i[n-1][iperiod - 1] * (*model).sm[ism] * green_day_i[ism][igeo][iperiod];
              } /* for mgeo */
          } /* for msm */
      for (igeo=0;igeo<MAX_S_COEFF;igeo++) geo_ind_field[igeo] += induc_sm[igeo];
    } /* if sm induced */

                                /* EST/IST field */

  if (fabs(est-9999.0) > 1.0e-10)
    {
      if (control[POS_EST_FAC] > 0) fac = (*model).est_fac; else fac = 1.0;

      if (control[POS_EST] > 0)
        {
          sm_est_field[0] = -fac * est;
          sm_est_field[1] = 0.0;
          sm_est_field[2] = 0.0;
          sm2geo(fday, 1, (*model).sm2geo_deg_ext, (*model).sm2geo_deg_day, (*model).coeff_sm2geo,
                 sm_est_field,est_in_geo);
          for (igeo=0;igeo<3;igeo++)
            geo_ext_field[igeo] +=  est_in_geo[igeo];
        }
      if (control[POS_IST] > 0)
        {
          sm_ist_field[0] = -fac * ist;
          sm_ist_field[1] = 0.0;
          sm_ist_field[2] = 0.0;
          sm2geo(fday, 1, (*model).sm2geo_deg_ext, (*model).sm2geo_deg_day, (*model).coeff_sm2geo,
                 sm_ist_field,ist_in_geo);
          for (igeo=0;igeo<3;igeo++)
            geo_ind_field[igeo] +=  ist_in_geo[igeo];
        }
    }

                                /* GSM */
  if (control[POS_GSM_IND] > control[POS_IMFBY_IND])
    gsm2geo_deg = control[POS_GSM_IND]; 
  else gsm2geo_deg = control[POS_IMFBY_IND];

  gsm2geo_green(fday, gsm2geo_deg, (*model).gsm2geo_deg_ext, (*model).gsm2geo_deg_day, (*model).gsm2geo_deg_ann, (*model).coeff_gsm2geo,
                green_day_r, green_day_i, green_ann_r, green_ann_i);

  if (control[POS_GSM] > 0)
    {
      gsm2geo(fday, control[POS_GSM], (*model).gsm2geo_deg_ext, (*model).gsm2geo_deg_day, (*model).gsm2geo_deg_ann, (*model).coeff_gsm2geo,
              (*model).gsm,gsm_in_geo);
      for (igeo=0;igeo<control[POS_GSM]*(control[POS_GSM]+2);igeo++)
        geo_ext_field[igeo] +=  gsm_in_geo[igeo];
    }

                                /* Induced by GSM in GEO */

  if (control[POS_GSM_IND] > 0)
    {
      for (igeo=0;igeo<MAX_S_COEFF;igeo++) induc_gsm[igeo] = 0;
          
      for (n = 1;n<=control[POS_GSM_IND];n++)
        for (mgsm = -n;mgsm<=n;mgsm++)
          {
            igsm = indx(n,mgsm);
            for (mgeo = -n;mgeo<=n;mgeo++)
              {
                igeo = indx(n,mgeo);
                for (iperiod = 1; iperiod <= (*model).gsm2geo_deg_day; iperiod++)	  /*green[,,0] is constant term*/ 
                  induc_gsm[igeo] += qfac_day_r[n-1][iperiod - 1] * (*model).gsm[igsm] * green_day_r[igsm][igeo][iperiod]
                    - CONVENTION * qfac_day_i[n-1][iperiod - 1] * (*model).gsm[igsm] * green_day_i[igsm][igeo][iperiod];
                for (iperiod = 1; iperiod <= (*model).gsm2geo_deg_ann; iperiod++)    /*green[,,0] is constant term*/ 
                  induc_gsm[igeo] +=   qfac_ann_r[n-1][iperiod - 1] * (*model).gsm[igsm] * green_ann_r[igsm][igeo][iperiod]
                    - CONVENTION * qfac_ann_i[n-1][iperiod - 1] * (*model).gsm[igsm] * green_ann_i[igsm][igeo][iperiod];
              } /* for mgeo */
          } /* for mgsm */
      for (igeo=0;igeo<MAX_S_COEFF;igeo++) geo_ind_field[igeo] += induc_gsm[igeo];
    } /* if gsm induced */


                                /* IMF-By */
  for (igsm=0;igsm<control[POS_IMFBY]*(control[POS_IMFBY]+2);igsm++)
    gsm_imfby_field[igsm] =  imf_by*(*model).imfby[igsm];

  if (control[POS_IMFBY] > 0)
    {
      gsm2geo(fday, control[POS_IMFBY], (*model).gsm2geo_deg_ext, (*model).gsm2geo_deg_day, (*model).gsm2geo_deg_ann, (*model).coeff_gsm2geo,
              gsm_imfby_field,imfby_in_geo);
      for (igeo=0;igeo<control[POS_IMFBY]*(control[POS_IMFBY]+2);igeo++)
        geo_ext_field[igeo] +=  imfby_in_geo[igeo];
    }


                                /* IMF-By induced */
  if (control[POS_IMFBY_IND] > 0)
    {
      for (igeo=0;igeo<MAX_S_COEFF;igeo++) induc_imfby[igeo] = 0;

      for (n = 1;n<=control[POS_IMFBY_IND];n++)
        for (mgsm = -n;mgsm<=n;mgsm++)
          {
            igsm = indx(n,mgsm);
            for (mgeo = -n;mgeo<=n;mgeo++)
              {
                igeo = indx(n,mgeo);
                for (iperiod = 1; iperiod <= (*model).gsm2geo_deg_day; iperiod++) 
                  induc_imfby[igeo] += (*model).imfby[igsm] * imf_by * (qfac_day_r[n-1][iperiod - 1] * green_day_r[igsm][igeo][iperiod]
                                                                        - CONVENTION * qfac_day_i[n-1][iperiod - 1] * green_day_i[igsm][igeo][iperiod]);
              } /* for mgeo */
          } /* for mgsm */
      for (igeo=0;igeo<MAX_S_COEFF;igeo++) geo_ind_field[igeo] += induc_imfby[igeo];
    }

  costh = cos(theta);
  rad = rrel*6371.2;
  plmbar_(&p[0],&dp[0],&costh,&ext_deg,&inorm);
  magfdz_(&p[0],&dp[0],&theta,&phi,&rad,&ext_deg,&geo_ind_field[0],
          &dx[0],&dy[0],&dz[0],&dpot[0],&xi,&yi,&zi,&h,&f,&ainc,&d,&pot);
  magfdz_ext_(&p[0],&dp[0],&theta,&phi,&rad,&ext_deg,&geo_ext_field[0],
          &dx[0],&dy[0],&dz[0],&dpot[0],&xe,&ye,&ze,&h,&f,&ainc,&d,&pot);
  *x = xe + xi;
  *y = ye + yi;
  *z = ze + zi;

  if (TEST) 
    {
      magfdz_ext_(&p[0],&dp[0],&theta,&phi,&rad,&ext_deg,&sm_in_geo[0],
          &dx[0],&dy[0],&dz[0],&dpot[0],&tx,&ty,&tz,&h,&f,&ainc,&d,&pot);
      printf("SM extern:  %10.4f %10.4f %10.4f\n",tx,ty,tz);
      magfdz_(&p[0],&dp[0],&theta,&phi,&rad,&ext_deg,&induc_sm[0],
          &dx[0],&dy[0],&dz[0],&dpot[0],&tx,&ty,&tz,&h,&f,&ainc,&d,&pot);
      printf("SM induced: %10.4f %10.4f %10.4f\n",tx,ty,tz);
      magfdz_ext_(&p[0],&dp[0],&theta,&phi,&rad,&ext_deg,&est_in_geo[0],
          &dx[0],&dy[0],&dz[0],&dpot[0],&tx,&ty,&tz,&h,&f,&ainc,&d,&pot);
      printf("SM Est:     %10.4f %10.4f %10.4f\n",tx,ty,tz);
      magfdz_(&p[0],&dp[0],&theta,&phi,&rad,&ext_deg,&ist_in_geo[0],
          &dx[0],&dy[0],&dz[0],&dpot[0],&tx,&ty,&tz,&h,&f,&ainc,&d,&pot);
      printf("SM Ist:     %10.4f %10.4f %10.4f\n",tx,ty,tz);

      magfdz_ext_(&p[0],&dp[0],&theta,&phi,&rad,&ext_deg,&gsm_in_geo[0],
          &dx[0],&dy[0],&dz[0],&dpot[0],&tx,&ty,&tz,&h,&f,&ainc,&d,&pot);
      printf("GSM extern: %10.4f %10.4f %10.4f\n",tx,ty,tz);
      magfdz_(&p[0],&dp[0],&theta,&phi,&rad,&ext_deg,&induc_gsm[0],
          &dx[0],&dy[0],&dz[0],&dpot[0],&tx,&ty,&tz,&h,&f,&ainc,&d,&pot);
      printf("GSM ind:    %10.4f %10.4f %10.4f\n",tx,ty,tz);
      magfdz_ext_(&p[0],&dp[0],&theta,&phi,&rad,&ext_deg,&imfby_in_geo[0],
          &dx[0],&dy[0],&dz[0],&dpot[0],&tx,&ty,&tz,&h,&f,&ainc,&d,&pot);
      printf("GSM IMFBy:  %10.4f %10.4f %10.4f\n",tx,ty,tz);
      magfdz_(&p[0],&dp[0],&theta,&phi,&rad,&ext_deg,&induc_imfby[0],
          &dx[0],&dy[0],&dz[0],&dpot[0],&tx,&ty,&tz,&h,&f,&ainc,&d,&pot);
      printf("IMFBy ind:  %10.4f %10.4f %10.4f\n",tx,ty,tz);
      
      printf("\nComplete result: %10.4f %10.4f %10.4f\n",*x,*y,*z);
      printf("<RETURN> ");
      getchar();
    }
  return;
} /* mainfield_external_field */

void mainfield_mp (int control[MAXCONTROL], tmodel *model,
                  double fday, double r, double theta, double phi,
                  double est, double ist, double imf_by,
                  double *x, double *y, double *z, double *f)
{
  double dx,dy,dz,rrel;
  mainfield_internal_field_mp (control, model, fday, r, theta, phi, x, y, z, f);

  rrel = r/6371.2;
  mainfield_external_field_mp (control, model, fday, rrel, theta, phi, est, ist, imf_by, &dx, &dy, &dz);
  *x += dx;
  *y += dy;
  *z += dz;
  *f = sqrt(*x**x + *y**y + *z**z);
  return;
} /* mainfield_mp */

void mainfield_geodetic_mp (int control[MAXCONTROL], tmodel *model,
                            double fday, double geod_alt, double geod_lat, double lon,
                            double est, double ist, double imf_by,
                            double *geod_x, double *geod_y, double *geod_z, double *h, double *f, double *incl, double *decl)
{
  double theta, phi, lat_rad, delta, dx,dy,dz, x,y,z, r, rrel;

  geodetic2geocentric(geod_lat*M_PI/180.0, geod_alt, &lat_rad, &r); /* input and output latitudes in radian*/

  theta = M_PI/2.0 - lat_rad;
  phi = lon * M_PI / 180.0;
  mainfield_internal_field_mp (control, model, fday, r, theta, phi, &x, &y, &z, f);

  rrel = r/6371.2;
  mainfield_external_field_mp (control, model, fday, rrel, theta, phi, est, ist, imf_by, &dx, &dy, &dz);
  x += dx;
  y += dy;
  z += dz;
  *f = sqrt(x*x + y*y + z*z);

  delta = (90.0-geod_lat) * M_PI / 180.0; /* delta is the geodetic co-latitude */
  geocentric2geodetic_vec(theta, delta, x, y, z, geod_x, geod_y, geod_z);

  *h     = sqrt(*geod_x * *geod_x + *geod_y * *geod_y);
  *incl  = 0.0;
  if (*f > 0.0) *incl = asin(*geod_z / *f);
  *incl *= 180.0 / M_PI;
  *decl  = atan2(*geod_y, *geod_x);
  *decl *= 180.0 / M_PI;
  return;
} /* mainfield_geodetic_mp */

